# React Finite State Machine

[Документация на английском](README.md)

Реакт компонент, реализующий логику конечного автомата.
Предоставляет возможность описывать состояния компонента, логику переходов между ними, передавать полезные данные между стадиями.

### Prop Types

#### FSM Component Props

| Property | Type | Required? | Description |
|:---|:---|:---:|:---|
| meta | Meta | ✓ | Метаданные. Описание стадий. |
| onFinish | OnFinish |  | Хук, срабатывающий после перехода в конечную стадию, вызова метода finish() в компоненте. |
| beforeTransition | BeforeTransition |  | Хук, срабатыающий перед переходом на заданную стадию. Возможно прервать переход на заданную стадию вернув false или Promise<false>. |
| afterTransition | AfterTransition |  | Хук, срабатывающий после перехода на заданную стадию. |
| Layout | React.ReactType<LayoutComponentProps> | | Общая разметка для всех узлов конечного автомата, которая получает информацию об актуальном шаге и children для отрисовки |
| commonProps | any | | Общие параметры, которые будут переданы в каждый компонент узла конченого автомата |

```js
type StageName = string | number;

/**
 * История переходов по стадиям
 */
interface IHistory {
	/** Количество записей в истории */
	recordsCount: number;

	/** Ссылка на объект первой записи в истории */
	first: IStage | null;

	/** Ссылка на объект последней записи в истории */
	last: IStage | null;

	/**
	 * Метод добавления новой записи в историю
	 * @param {StageName} stageName Stage name
	 * @param {any} payload
	 * @returns {void}
	 */
	add: (stageName: StageName, payload: any) => void;
}

type Meta<Stage extends StageName = string> = {
	/** Список стадий и соответствующих компонент (пара 'name-component') */
	stages: {
		[name in Stage]: React.ReactType;
	};

	/** Начальная стадия */
	initialStage: Stage;

	/** Конечная стадия */
	finalStage: Stage;
};

type OnFinish: (transitionsHistory: IHistory) => void;

type BeforeTransition = (currentStage: StageName, nextStage: StageName, payload?: any) => boolean;

type AfterTransition = (currentStage: StageName, prevStage: StageName, payload?: any) => void;

type CommonProps = any;

type Layout = React.ReactType<LayoutComponentProps>;
```

#### Stage Component Props

| Property | Type | Description |
|:---|:---|:---|
| transition | Transition | Метод перехода к заданной стадии. Может принимать полезные данные для передачи на следующую стадию. |
| toPrevious | ToPrevious | Метод возврата к предыдущей стадии. Может принимать полезные данные для передачи на предыдущю стадию. |
| finish | () => void | Вызывает onFinish хук. |
| payload | IncomingPayload | Данные, переданные с предыдущей стадии. |
| commonProps | any | Общие параметры, которые будут переданы в каждый компонент узла конченого автомата |

```js
type IncomingPayload = {} | void;

type OutGoingPayload = {} | void;

type ToPrevPayload = OutGoingPayload;

type Transition = (stageName: StageName, payload?: OutGoingPayload) => Promise<void>;

type ToPrevious = (payload?: ToPrevPayload) => Promise<void>;

type CommonProps = any;
```

### Examples
```js
import React from 'react';
import ReactDOM from 'react-dom';
import {FSM} from 'react-fsm';

const commonProps = {
	key: 'value'
};

const Layout = ({children, currentStage, currentStagePayload}) => (
	<div>
		{children}
	</div>
);

const Stage = {
	Stage1: 'Stage1',
	Stage2: 'Stage2',
	Stage3: 'Stage3'
};

const formsMeta = {
	stages: {
		[Stage.Stage1]: ({transition}) => (
			<div onClick={() => transition(Stage.Stage2)}>
				Stage 1
			</div>
		),
		[Stage.Stage2]: ({transition}) => (
			<div onClick={() => transition(Stage.Stage3)}>
				Stage 2
			</div>
		),
		[Stage.Stage3]: ({finish}) => (
			<div onClick={() => finish()}>
				Stage 3
			</div>
		)
	},
	initialStage: Stage.Stage1,
	finalStage: Stage.Stage3
};

const Form = () => (
	<FSM
		meta={formsMeta}
		afterTransition={(currentStage, prevStage, payload) => {
			console.log(
				`We've made a transition to the stage ${currentStage}`,
				`From stage ${prevStage}`,
				`Received payload: ${payload}`
			);
		}}
		beforeTransition={(currentStage, nextStage, payload) => {
			console.log(
				`Try to make a transition to the stage ${nextStage}`,
				`From stage ${prevStage}`,
				`With payload: ${payload}`
			);

			return !!payload.hasCats;
		}}
		onFinish={(transitionsHistory) => {
			console.log(`Finish. History: ${transitionsHistory}`);
		}}
		commonProps={commonProps}
		Layout={Layout}
	/>
);

ReactDOM.render(
  <Form />,
  document.getElementById('example')
);
```
