import * as React from 'react';

export interface ISerializable {
	/** Returns the json representation of an object */
	toJSON(): object;
}

export type StageName = string | number;

/** Stage */
export interface IStage extends ISerializable {
	/** Stage name */
	name: StageName;

	/** Stage's payload */
	payload: any;

	/** Reference to the previous stage */
	prevStage: IStage | null;

	/** Reference to the next stage */
	nextStage: IStage | null;
}

/** History of transitions */
export interface IHistory extends ISerializable {
	/** Number of history records */
	recordsCount: number;

	/** Reference to the first record */
	first: IStage | null;

	/** Reference to the last record */
	last: IStage | null;

	/**
	 * Adds a new record to the history
	 * @param {StageName} stageName Stage name
	 * @param {any} payload
	 * @returns {void}
	 */
	add: (stageName: StageName, payload: any) => void;
}

/** Props passed to stage component */
export type StageComponentProps<
	IncomingPayload = {} | void,
	OutGoingPayload = {} | void,
	ToPrevPayload = OutGoingPayload,
	CommonProps = {}
> = {
	/**
	 * Transitions to the stage
	 * @param {StageName} stageName Stage name
	 * @param {OutGoingPayload} [payload] Payload for the next stage
	 * @returns {Promise<void>}
	 */
	transition: (stageName: StageName, payload?: OutGoingPayload) => Promise<void>;

	/**
	 * Returns to previous stage
	 * @param {ToPrevPayload} payload Payload for the next stage
	 * @returns {Promise<void>}
	 */
	toPrevious: (payload?: ToPrevPayload) => Promise<void>;

	/**
	 * Call onFinish hook
	 * @returns {void}
	 */
	finish: () => void;

	/** Payload passed from previous stage */
	payload: IncomingPayload;

	/**
	 *  General parameters for all nodes of a finite state machine
	 */
	commonProps: CommonProps;
};

export type StageComponent = React.ElementType<StageComponentProps>;

export type LayoutComponentProps = {
	currentStage: StageName;
	currentStagePayload?: any;
};

export type Meta<Stage extends StageName = string> = {
	/** List of stages - 'name-component' pairs */
	stages: {
		[name in Stage]: React.ElementType<StageComponentProps>;
	};

	/** Initial stage for the state machine */
	initialStage: Stage;
};

/** Before transition hook */
export type BeforeTransition = (currentStage: StageName, nextStage: StageName, payload?: any) => boolean;

/** After transition hook */
export type AfterTransition = (currentStage: StageName, prevStage: StageName, payload?: any) => void;

/** On finish hook */
export type OnFinish = (transitionsHistory: IHistory) => void;

export type FSMProps = {
	meta: Meta;
	onFinish?: OnFinish;
	beforeTransition?: BeforeTransition;
	afterTransition?: AfterTransition;
	/**
	 * General layout for all nodes of a finite state machine that receives
	 * information about the current step and children for rendering
	 */
	Layout?: React.ElementType<LayoutComponentProps>;
	/**
	 * General parameters that will be sent to each node component of a finite state machine
	 */
	commonProps?: any;
};

export declare const FSM: React.FC<FSMProps>;
